"""TSP Art Module for Python"""

__version__ = "0.6.17"


import tspart.voronoi
import tspart.tsp
import tspart.neos
import tspart.studio

from tspart._draw import draw_route, draw_cmyk_routes, draw_rgb_routes
from tspart._image import split_rgb, split_cmyk, rgb_to_grayscale, image_to_base64, base64_to_image
from tspart._helpers import (
    image_to_array, array_to_image, map_points_to_tour, map_points_to_tour_multi, image_array_size, factors_from_image,
    factors_from_image_multi, filter_white_points, filter_white_points_multi
)
from tspart._files import (
    save_array_as_image, load_image_as_array, make_tsplib, decode_tsplib, save_tsplib, load_tsplib, save_cyc_tour,
    load_cyc_tour
)



from typing import Tuple, Sequence

import numpy as np
from PIL import Image
import scipy.spatial
import scipy.ndimage


def get_bounding_corners(
        points: Sequence[Sequence[float]]
) -> Tuple[float]:
    points = np.array(points)

    start = np.floor(points.min(0)).astype(int)
    end = np.ceil(points.max(0)).astype(int)

    return start, end


def image_to_array(
        image: Image,
        mode: str = "RGB"
) -> np.ndarray[np.ndarray[float | np.ndarray[float]]]:
    return np.array(image.convert(mode))


def array_to_image(
        array: np.ndarray[np.ndarray[float | np.ndarray[float]]],
        mode: str = "RGB"
) -> Image:
    return Image.fromarray(array, mode)


def map_points_to_tour(
        points: Sequence[Sequence[float]],
        tour: Sequence[int]
) -> np.ndarray[np.ndarray[float]]:
    return np.array([points[idx] for idx in tour])


def map_points_to_tour_multi(
        points_list: Sequence[Sequence[Sequence[float]]],
        tours_list: Sequence[Sequence[int]]
) -> list[np.ndarray[np.ndarray[float]]]:
    result = []
    for idx, points in enumerate(points_list):
        routes = tours_list[idx]

        r = map_points_to_tour(
            points,
            routes
        )

        result.append(r)

    return result


def ndarray_to_array_2d(array, round_places=None):
    if round_places is not None:
        return [list([list(__.round(round_places)) for __ in _]) for _ in array]

    return [list([list(__) for __ in _]) for _ in array]


def array_to_ndarray_2d(array, round_places=None):
    if round_places is not None:
        return [np.array(_).round(round_places) for _ in array]

    return [np.array(_) for _ in array]


def nearest_point_index(points, location):
    distance, index = scipy.spatial.KDTree(points).query(location)

    return int(index)


def image_array_size(img):
    return tuple([int(round(_)) for _ in img.shape[:2]][::-1])


def line_angle(start, end):
    d = np.array(end) - np.array(start)

    return np.arctan2(d[1], d[0])


def circle_point(radius, angle, offset=(0, 0)):
    offset = np.array(offset)

    x = radius * np.cos(angle)
    y = radius * np.sin(angle)

    point = np.array([x, y])

    return point + offset


def get_point_value(grayscale_array, point):
    y, x = np.floor(np.array(point)).astype(int)

    return grayscale_array[x][y]


def size_factor(grayscale_array, point):
    value = get_point_value(
        grayscale_array=grayscale_array,
        point=point
    )
    return 1 - (value / 255)


def factors_from_image(grayscale_array, points, blur_sigma=1):
    if blur_sigma > 0:
        grayscale_array = scipy.ndimage.gaussian_filter(grayscale_array, sigma=blur_sigma)

    factors = []
    for point in points:
        factor = size_factor(
            grayscale_array=grayscale_array,
            point=point
        )
        factors.append(factor)

    return np.array(factors)


def factors_from_image_multi(grayscale_arrays, points_list, blur_sigma=1):
    factors_list = []
    for grayscale_array, points in zip(grayscale_arrays, points_list):
        factors = factors_from_image(
            grayscale_array=grayscale_array,
            points=points,
            blur_sigma=blur_sigma
        )

        factors_list.append(factors)

    return factors_list


def filter_white_points(grayscale_array, points, threshold=254, blur_sigma=1):
    if blur_sigma > 0:
        grayscale_array = scipy.ndimage.gaussian_filter(grayscale_array, sigma=blur_sigma)

    result = []
    for point in points:
        value = get_point_value(
            grayscale_array=grayscale_array,
            point=point
        )
        if value <= threshold:
            result.append(point)
    return result


def filter_white_points_multi(grayscale_arrays, points_list, threshold=1, blur_sigma=1):
    results = []
    for grayscale_array, points in zip(grayscale_arrays, points_list):
        points = filter_white_points(
            grayscale_array=grayscale_array,
            points=points,
            threshold=threshold,
            blur_sigma=blur_sigma
        )

        results.append(points)

    return results



# -----------------------------------------------------------------------------
# Weighted Voronoi Stippler
# Copyright (2017) Nicolas P. Rougier - BSD license
# -----------------------------------------------------------------------------
import numpy as np
import scipy.spatial


def rasterize(V):
    """
    Polygon rasterization (scanlines).

    Given an ordered set of vertices V describing a polygon,
    return all the (integer) points inside the polygon.
    See http://alienryderflex.com/polygon_fill/

    Parameters:
    -----------

    V : (n,2) shaped numpy array
        Polygon vertices
    """

    n = len(V)
    X, Y = V[:, 0], V[:, 1]
    ymin = int(np.ceil(Y.min()))
    ymax = int(np.floor(Y.max()))
    # ymin = int(np.round(Y.min()))
    # ymax = int(np.round(Y.max()))
    P = []
    for y in range(ymin, ymax + 1):
        segments = []
        for i in range(n):
            index1, index2 = (i - 1) % n, i
            y1, y2 = Y[index1], Y[index2]
            x1, x2 = X[index1], X[index2]
            if y1 > y2:
                y1, y2 = y2, y1
                x1, x2 = x2, x1
            elif y1 == y2:
                continue
            if (y1 <= y < y2) or (y == ymax and y1 < y <= y2):
                segments.append((y - y1) * (x2 - x1) / (y2 - y1) + x1)

        segments.sort()
        for i in range(0, (2 * (len(segments) // 2)), 2):
            x1 = int(np.ceil(segments[i]))
            x2 = int(np.floor(segments[i + 1]))
            # x1 = int(np.round(segments[i]))
            # x2 = int(np.round(segments[i+1]))
            P.extend([[x, y] for x in range(x1, x2 + 1)])
    if not len(P):
        return V
    return np.array(P)


def rasterize_outline(V):
    """
    Polygon outline rasterization (scanlines).

    Given an ordered set of vertices V describing a polygon,
    return all the (integer) points for the polygon outline.
    See http://alienryderflex.com/polygon_fill/

    Parameters:
    -----------

    V : (n,2) shaped numpy array
        Polygon vertices
    """
    n = len(V)
    X, Y = V[:, 0], V[:, 1]
    ymin = int(np.ceil(Y.min()))
    ymax = int(np.floor(Y.max()))
    points = np.zeros((2 + (ymax - ymin) * 2, 3), dtype=int)
    index = 0
    for y in range(ymin, ymax + 1):
        segments = []
        for i in range(n):
            index1, index2 = (i - 1) % n, i
            y1, y2 = Y[index1], Y[index2]
            x1, x2 = X[index1], X[index2]
            if y1 > y2:
                y1, y2 = y2, y1
                x1, x2 = x2, x1
            elif y1 == y2:
                continue
            if (y1 <= y < y2) or (y == ymax and y1 < y <= y2):
                segments.append((y - y1) * (x2 - x1) / (y2 - y1) + x1)
        segments.sort()
        for i in range(0, (2 * (len(segments) // 2)), 2):
            x1 = int(np.ceil(segments[i]))
            x2 = int(np.ceil(segments[i + 1]))
            points[index] = x1, x2, y
            index += 1
    return points[:index]


def weighted_centroid_outline(V, P, Q):
    """
    Given an ordered set of vertices V describing a polygon,
    return the surface weighted centroid according to density P & Q.

    P & Q are computed relatively to density:
    density_P = density.cumsum(axis=1)
    density_Q = density_P.cumsum(axis=1)

    This works by first rasterizing the polygon and then
    finding the center of mass over all the rasterized points.
    """

    O = rasterize_outline(V)
    X1, X2, Y = O[:, 0], O[:, 1], O[:, 2]

    Y = np.minimum(Y, P.shape[0] - 1)
    X1 = np.minimum(X1, P.shape[1] - 1)
    X2 = np.minimum(X2, P.shape[1] - 1)

    d = (P[Y, X2] - P[Y, X1]).sum()
    x = ((X2 * P[Y, X2] - Q[Y, X2]) - (X1 * P[Y, X1] - Q[Y, X1])).sum()
    y = (Y * (P[Y, X2] - P[Y, X1])).sum()
    if d:
        return [x / d, y / d]
    return [x, y]


def uniform_centroid(V):
    """
    Given an ordered set of vertices V describing a polygon,
    returns the uniform surface centroid.

    See http://paulbourke.net/geometry/polygonmesh/
    """
    A = 0
    Cx = 0
    Cy = 0
    for i in range(len(V) - 1):
        s = (V[i, 0] * V[i + 1, 1] - V[i + 1, 0] * V[i, 1])
        A += s
        Cx += (V[i, 0] + V[i + 1, 0]) * s
        Cy += (V[i, 1] + V[i + 1, 1]) * s
    Cx /= 3 * A
    Cy /= 3 * A
    return [Cx, Cy]


def weighted_centroid(V, D):
    """
    Given an ordered set of vertices V describing a polygon,
    return the surface weighted centroid according to density D.

    This works by first rasterizing the polygon and then
    finding the center of mass over all the rasterized points.
    """

    P = rasterize(V)
    Pi = P.astype(int)
    Pi[:, 0] = np.minimum(Pi[:, 0], D.shape[1] - 1)
    Pi[:, 1] = np.minimum(Pi[:, 1], D.shape[0] - 1)
    D = D[Pi[:, 1], Pi[:, 0]].reshape(len(Pi), 1)
    return ((P * D)).sum(axis=0) / D.sum()


# http://stackoverflow.com/questions/28665491/...
#    ...getting-a-bounded-polygon-coordinates-from-voronoi-cells
def in_box(points, bbox):
    return np.logical_and(
        np.logical_and(bbox[0] <= points[:, 0], points[:, 0] <= bbox[1]),
        np.logical_and(bbox[2] <= points[:, 1], points[:, 1] <= bbox[3]))


def voronoi(points, bbox):
    # See http://stackoverflow.com/questions/28665491/...
    #   ...getting-a-bounded-polygon-coordinates-from-voronoi-cells
    # See also https://gist.github.com/pv/8036995

    # Select points inside the bounding box
    i = in_box(points, bbox)

    # Mirror points
    points_center = points[i, :]
    points_left = np.copy(points_center)
    points_left[:, 0] = bbox[0] - (points_left[:, 0] - bbox[0])
    points_right = np.copy(points_center)
    points_right[:, 0] = bbox[1] + (bbox[1] - points_right[:, 0])
    points_down = np.copy(points_center)
    points_down[:, 1] = bbox[2] - (points_down[:, 1] - bbox[2])
    points_up = np.copy(points_center)
    points_up[:, 1] = bbox[3] + (bbox[3] - points_up[:, 1])
    points = np.append(points_center,
                       np.append(np.append(points_left, points_right, axis=0),
                                 np.append(points_down, points_up, axis=0),
                                 axis=0), axis=0)
    # Compute Voronoi
    vor = scipy.spatial.Voronoi(points)

    # Filter regions
    epsilon = 0.1
    regions = []
    for region in vor.regions:
        flag = True
        for index in region:
            if index == -1:
                flag = False
                break
            else:
                x = vor.vertices[index, 0]
                y = vor.vertices[index, 1]
                if not (bbox[0] - epsilon <= x <= bbox[1] + epsilon and
                        bbox[2] - epsilon <= y <= bbox[3] + epsilon):
                    flag = False
                    break
        if region != [] and flag:
            regions.append(region)
    vor.filtered_points = points_center
    vor.filtered_regions = regions
    return vor


def centroids(points, density, density_P=None, density_Q=None):
    """
    Given a set of point and a density array, return the set of weighted
    centroids.
    """

    X, Y = points[:, 0], points[:, 1]
    # You must ensure:
    #   0 < X.min() < X.max() < density.shape[0]
    #   0 < Y.min() < Y.max() < density.shape[1]

    xmin, xmax = 0, density.shape[1]
    ymin, ymax = 0, density.shape[0]
    bbox = np.array([xmin, xmax, ymin, ymax])
    vor = voronoi(points, bbox)
    regions = vor.filtered_regions
    centroids = []
    for region in regions:
        vertices = vor.vertices[region + [region[0]], :]
        # vertices = vor.filtered_points[region + [region[0]], :]

        # Full version from all the points
        # centroid = weighted_centroid(vertices, density)

        # Optimized version from only the outline
        centroid = weighted_centroid_outline(vertices, density_P, density_Q)

        centroids.append(centroid)
    return regions, np.array(centroids)

# ---- END OF ORIGINAL FILE ---- #
import sys
import scipy.ndimage

from etatime import EtaBar


def normalize(D):
    Vmin, Vmax = D.min(), D.max()
    if Vmax - Vmin > 1e-5:
        D = (D - Vmin) / (Vmax - Vmin)
    else:
        D = np.zeros_like(D)
    return D


def initialization(n, D):
    """
    Return n points distributed over [xmin, xmax] x [ymin, ymax]
    according to (normalized) density distribution.

    with xmin, xmax = 0, density.shape[1]
         ymin, ymax = 0, density.shape[0]

    The algorithm here is a simple rejection sampling.
    """

    samples = []
    while len(samples) < n:
        # X = np.random.randint(0, D.shape[1], 10*n)
        # Y = np.random.randint(0, D.shape[0], 10*n)
        X = np.random.uniform(0, D.shape[1], 10 * n)
        Y = np.random.uniform(0, D.shape[0], 10 * n)
        P = np.random.uniform(0, 1, 10 * n)
        index = 0
        while index < len(X) and len(samples) < n:
            x, y = X[index], Y[index]
            x_, y_ = int(np.floor(x)), int(np.floor(y))
            if P[index] < D[y_, x_]:
                samples.append([x, y])
            index += 1
    return np.array(samples)


def stipple_image(grayscale_array, points=5000, iterations=50, logging=True):
    # We want (approximately) 500 pixels per voronoi region
    zoom = (points * 500) / (grayscale_array.shape[0] * grayscale_array.shape[1])
    density = scipy.ndimage.zoom(grayscale_array, zoom, order=0)

    density = 1.0 - normalize(density)
    density_P = density.cumsum(axis=1)
    density_Q = density_P.cumsum(axis=1)

    # Initialization
    stippled = initialization(points, density)

    if logging:
        iter = EtaBar(range(iterations))
    else:
        iter = range(iterations)

    for _ in iter:
        regions, stippled = centroids(stippled, density, density_P, density_Q)

    return stippled / zoom


def stipple_image_multi(grayscale_arrays, points=5000, iterations=50, logging=True):
    result = []
    for idx, grayscale_array in enumerate(grayscale_arrays):
        if logging:
            print(f"Stippling image {idx + 1}/{len(grayscale_arrays)}", file=sys.stderr)

        stippled = stipple_image(
            grayscale_array=grayscale_array,
            points=points,
            iterations=iterations,
            logging=logging
        )

        result.append(stippled)

    return result



import sys
import time
import xmlrpc.client

from tspart._files import make_tsplib as _make_tsplib
from tspart._helpers import map_points_to_tour as _map_points_to_tour


class NeosPingError(ConnectionError):
    pass


class NeosSubmitError(ConnectionError):
    pass


class NeosSolveError(RuntimeError):
    pass


def get_client(url="https://neos-server.org:3333"):
    client = xmlrpc.client.ServerProxy(url)

    if client.ping() != "NeosServer is alive\n":
        raise NeosPingError("Could not verify that Neos is online")

    return client


def make_solver_job(email, points):
    result = "<document>\n"

    result += "<category>co</category>\n"
    result += "<solver>concorde</solver>\n"
    result += "<inputMethod>TSP</inputMethod>\n"
    result += f"<email><![CDATA[{email}]]></email>\n"

    result += "<tsp><![CDATA[\n"
    result += _make_tsplib(points).replace("\r\n", "\n")
    result += "]]></tsp>\n"

    result += "<ALGTYPE><![CDATA[con]]></ALGTYPE>\n"
    result += "<RDTYPE><![CDATA[fixed]]></RDTYPE>\n"
    result += "<PLTYPE><![CDATA[no]]></PLTYPE>\n"

    result += "<comment><![CDATA[]]></comment>\n"

    result += "</document>"

    return result


def submit_solve(client, email, points):
    xml_request = make_solver_job(email, points)

    job_number, password = client.submitJob(xml_request)

    if job_number == 0:
        raise NeosSubmitError(password)

    return job_number, password


def submit_solves(client, email, points_list):
    result = []
    for points in points_list:
        r = submit_solve(
            client=client,
            email=email,
            points=points
        )

        result.append(r)

    return result


def cancel_solve(client, job_number, password):
    client.killJob(job_number, password)


def cancel_solves(client, job_list):
    for job_number, password in job_list:
        cancel_solve(
            client=client,
            job_number=job_number,
            password=password
        )


def get_solve(client, job_number, password, points=None):
    if client.getJobStatus(job_number, password) != "Done":
        return None

    completion_code = client.getCompletionCode(job_number, password)
    neos_results = client.getFinalResults(job_number, password)
    neos_results = neos_results.data.decode()
    if completion_code != "Normal":
        raise NeosSolveError(f"Neos job completed with failed code: {completion_code}\n\n"
                             f"Full response:\n"
                             f"{neos_results}")

    results_arr = [_.strip() for _ in neos_results.split("\n") if _.strip() != ""]

    # Get the tour indexes from the raw response
    raw_tour = []
    for line in results_arr[::-1]:
        if not all([_.isnumeric() for _ in line.split(" ")]):
            break

        raw_tour.append(line)

    if len(raw_tour) == 0:
        raise NeosSolveError(f"Neos job did not return any data, got response:\n\n{neos_results}")

    raw_tour = raw_tour[::-1][1:]

    if len(raw_tour[0].split(" ")) == 3:
        short_mode = True
    else:
        short_mode = False

    tour = []
    for line in raw_tour:
        line = line.split(" ")
        if short_mode:
            tour.append(int(line[0]))
        else:
            nums = [int(_) for _ in line]
            tour += nums

    if points is None:
        return tour

    return _map_points_to_tour(points, tour)


def get_solves(client, job_list, points_list=None):
    result = []
    for idx, (job_number, password) in enumerate(job_list):
        points = None
        if points_list is not None:
            points = points_list[idx]

        r = get_solve(
            client=client,
            job_number=job_number,
            password=password,
            points=points
        )

        result.append(r)

    return result


def get_solve_blocking(client, job_number, password, points=None, delay_minutes=0.25, logging=True):
    result = None

    while result is None:
        result = get_solve(
            client=client,
            job_number=job_number,
            password=password,
            points=points
        )
        if logging:
            print("Still waiting for solve...", file=sys.stderr)

        time.sleep(delay_minutes * 60)

    return result


def get_solves_blocking(client, job_list, points_list=None, delay_minutes=0.25, logging=True):
    n = len(job_list)

    result = [None] * n
    results_not_done = [True] * n

    while any(results_not_done):
        for idx, (job_number, password) in enumerate(job_list):
            points = None
            if points_list is not None:
                points = points_list[idx]

            if result[idx] is None:
                result[idx] = get_solve(
                    client=client,
                    job_number=job_number,
                    password=password,
                    points=points
                )

        results_not_done = [_ is None for _ in result]
        num_solves = sum([not _ for _ in results_not_done])
        if logging:
            if num_solves < n:
                print(f"{num_solves}/{n} solves so far...", file=sys.stderr)
            else:
                print(f"{num_solves}/{n} solves done!", file=sys.stderr)

        if any(results_not_done):
            time.sleep(delay_minutes * 60)

    return result



import sys
import scipy.spatial
from ortools.constraint_solver import pywrapcp, routing_enums_pb2

from tspart._helpers import get_bounding_corners as _get_bounding_corners
from tspart._helpers import nearest_point_index as _nearest_point_index
from tspart._helpers import map_points_to_tour as _map_points_to_tour


# See https://developers.google.com/optimization/routing/tsp
def heuristic_solve(points, time_limit_minutes=60, symmetric=True, logging=True, verbose=False):
    distance_matrix = scipy.spatial.distance.cdist(points, points).round().astype(int)
    num_points = len(points)

    if symmetric:
        manager = pywrapcp.RoutingIndexManager(num_points, 1, 0)
    else:
        size = (_get_bounding_corners(points)[1] + 1)

        h_middle = int(round(size[0] / 2))
        start = _nearest_point_index(points, [h_middle, 0])
        end = _nearest_point_index(points, [h_middle, size[1]])

        manager = pywrapcp.RoutingIndexManager(num_points, 1, [start], [end])

    routing_parameters = pywrapcp.DefaultRoutingModelParameters()
    if logging and verbose:
        routing_parameters.solver_parameters.trace_propagation = True
        routing_parameters.solver_parameters.trace_search = True

    routing = pywrapcp.RoutingModel(manager, routing_parameters)

    def distance_callback(from_idx, to_idx):
        from_node = manager.IndexToNode(from_idx)
        to_node = manager.IndexToNode(to_idx)
        return distance_matrix[from_node][to_node]
    transit_callback_index = routing.RegisterTransitCallback(distance_callback)

    routing.SetArcCostEvaluatorOfAllVehicles(transit_callback_index)

    search_parameters = pywrapcp.DefaultRoutingSearchParameters()
    search_parameters.first_solution_strategy = (
        routing_enums_pb2.FirstSolutionStrategy.PATH_CHEAPEST_ARC
    )

    search_parameters.local_search_metaheuristic = (
        routing_enums_pb2.LocalSearchMetaheuristic.GUIDED_LOCAL_SEARCH)
    if time_limit_minutes:
        search_parameters.time_limit.seconds = int(round(time_limit_minutes * 60))

    if logging:
        search_parameters.log_search = True

    solution = routing.SolveWithParameters(search_parameters)

    if solution:
        index = routing.Start(0)
        tour = [manager.IndexToNode(index)]

        while not routing.IsEnd(index):
            index = solution.Value(routing.NextVar(index))
            tour.append(manager.IndexToNode(index))

        route = _map_points_to_tour(points, tour)

        return route
    else:
        return None


def heuristic_solves(points_list, time_limit_minutes=60, symmetric=True, logging=True, verbose=False):
    result = []
    for idx, points in enumerate(points_list):
        print(f"Solving image {idx + 1}/{len(points_list)}", file=sys.stderr)
        r = heuristic_solve(
            points=points,
            time_limit_minutes=time_limit_minutes,
            symmetric=symmetric,
            logging=logging,
            verbose=verbose
        )

        result.append(r)

    return result



import os
import shutil
import sys
import time
from enum import Enum
from typing import Sequence, Tuple

import numpy as np
from PIL import Image

from tspart._image import split_cmyk as _split_cmyk
from tspart._image import split_rgb as _split_rgb
from tspart._image import rgb_to_grayscale as _rgb_to_grayscale
from tspart._image import base64_to_image as _base64_to_image
from tspart._image import image_to_base64 as _image_to_base64
from tspart._draw import draw_cmyk_routes as _draw_cmyk_routes
from tspart._draw import draw_rgb_routes as _draw_rgb_routes
from tspart._draw import draw_route as _draw_route
from tspart._helpers import array_to_image as _array_to_image
from tspart._helpers import image_to_array as _image_to_array
from tspart._helpers import image_array_size as _image_array_size
from tspart._helpers import filter_white_points_multi as _filter_white_points_multi
from tspart._helpers import factors_from_image_multi as _factors_from_image_multi
from tspart._helpers import ndarray_to_array_2d as _ndarray_to_array_2d
from tspart._helpers import array_to_ndarray_2d as _array_to_ndarray_2d
from tspart._files import save_json as _save_json
from tspart._files import load_json as _load_json
from tspart import voronoi as _voronoi
from tspart import neos as _neos
from tspart import tsp as _tsp


class ColorMode(Enum):
    CMYK = "CMYK"
    RGB = "RGB"
    GRAYSCALE = "L"


class InadequateResultsWarning(RuntimeWarning):
    pass


class InadequateResultsError(RuntimeError):
    pass


class TspStudio:
    def __init__(
            self,
            mode: ColorMode,
            image: Image.Image | str,
            num_points: int = 5000,
            line_width: float = 2,
            white_threshold: int = 254,
            invert: bool = False,
            background: Tuple[int, int, int] | str = (255, 255, 255),
            foreground: Tuple[int, int, int] | str = (0, 0, 0),
            points: Sequence[Sequence[float | Sequence[float]]] | None = None,
            is_routed: bool = False,
            jobs: Sequence[Sequence[int | str]] | None = None
    ):
        self.neos = None
        self.factors = None

        self.mode = mode
        self.invert = invert
        self.image = image
        self.num_points = num_points
        self.line_width = line_width
        self.white_threshold = white_threshold

        self.background = background
        self.foreground = foreground
        self.points = points
        self.is_routed = is_routed
        self.jobs = jobs

    @property
    def mode(self) -> ColorMode:
        return self._mode

    @mode.setter
    def mode(self, value: ColorMode | str):
        if isinstance(value, ColorMode):
            self._mode = value
        else:
            self._mode = ColorMode(value)

    @property
    def image(self) -> Image:
        return _array_to_image(self._image)

    @image.setter
    def image(self, value: Image.Image | str):
        if isinstance(value, str):
            self.image_string = value
            self._image = _image_to_array(_base64_to_image(self.image_string))
        else:
            self._image = _image_to_array(value)
            self.image_string = _image_to_base64(_array_to_image(self._image))

        self.size = _image_array_size(self._image)

        match self.mode:
            case ColorMode.CMYK:
                self.channels = _split_cmyk(self._image, invert=self._real_invert)
            case ColorMode.RGB:
                self.channels = _split_rgb(self._image, invert=self._real_invert)
            case ColorMode.GRAYSCALE:
                self.channels = [_rgb_to_grayscale(self._image, invert=self._real_invert)]

    @property
    def num_points(self) -> int:
        return self._num_points

    @num_points.setter
    def num_points(self, value: int):
        self._num_points = max(1, value)

        self.points = None

    @property
    def line_width(self) -> float:
        return self._line_width

    @line_width.setter
    def line_width(self, value: float | int):
        self._line_width = max(0, value)

    @property
    def white_threshold(self) -> int:
        return self._white_threshold

    @white_threshold.setter
    def white_threshold(self, value: int):
        self._white_threshold = min(255, max(0, value))

    @property
    def invert(self):
        return self._invert

    @invert.setter
    def invert(self, value: bool):
        self._invert = value

        match self.mode:
            case ColorMode.CMYK:
                self._real_invert = self._invert
            case ColorMode.RGB:
                self._real_invert = not self._invert
            case ColorMode.GRAYSCALE:
                self._real_invert = self._invert

    @property
    def background(self):
        return self._background

    @background.setter
    def background(self, value):
        self._background = value

    @property
    def foreground(self):
        return self._foreground

    @foreground.setter
    def foreground(self, value):
        self._foreground = value

    @property
    def points(self) -> Sequence[Sequence[float | Sequence[float]]]:
        return self._points

    @points.setter
    def points(self, value: Sequence[Sequence[float | Sequence[float]]] | None):
        self._points = value
        if self._points is None:
            self.factors = None
        else:
            self.compute_factors()

        self.is_routed = False

    @property
    def is_routed(self) -> bool:
        return self._is_routed

    @is_routed.setter
    def is_routed(self, value: bool):
        self._is_routed = value

    @property
    def jobs(self) -> Sequence[Sequence[int | str]]:
        return self._jobs

    @jobs.setter
    def jobs(self, value: Sequence[Sequence[int | str]] | None):
        self._jobs = value

    @property
    def data(self):
        return {
            "mode": self.mode,
            "num_points": self.num_points,
            "line_width": self.line_width,
            "white_threshold": self.white_threshold,
            "invert": self.invert,
            "background": self.background,
            "foreground": self.foreground,
            "is_routed": self.is_routed,
            "jobs": self.jobs,
            "points": self.points,
            "image": self.image_string
        }

    def save(self, filename):
        save(
            filename=filename,
            studio=self
        )

    def stipple(self, iterations=50, logging=True):
        self.points = _voronoi.stipple_image_multi(
            grayscale_arrays=self.channels,
            points=self.num_points,
            iterations=iterations,
            logging=logging
        )

        if self.white_threshold < 255:
            self.points = _filter_white_points_multi(
                grayscale_arrays=self.channels,
                points_list=self.points,
                threshold=self.white_threshold
            )

        for idx, p in enumerate(self.points):
            length = len(p)
            if length == 0:
                raise InadequateResultsError(f"Channel {idx}: "
                                             f"Stippling produced no points, try raising the white threshold")
            if length < 100:
                raise InadequateResultsWarning(f"Channel {idx}: "
                                               f"Stippling produced very few ({length}) points, solves may fail")

    def compute_factors(self):
        if self.points is None:
            raise ValueError("Points not initialized")
        self.factors = _factors_from_image_multi(
            grayscale_arrays=self.channels,
            points_list=self.points
        )

    def setup_online_solves(self):
        if self.neos is None:
            self.neos = _neos.get_client()

    def submit_online_solve(self, points, email):
        self.setup_online_solves()

        return _neos.submit_solve(
            client=self.neos,
            email=email,
            points=points
        )

    def submit_online_solves(self, email):
        if self.points is None:
            raise ValueError("Points not initialized")

        self.setup_online_solves()
        self.cancel_online_solves()

        self.jobs = _neos.submit_solves(
            client=self.neos,
            email=email,
            points_list=self.points
        )

    def cancel_online_solves(self):
        self.setup_online_solves()

        if self.jobs is not None:
            _neos.cancel_solves(
                client=self.neos,
                job_list=self.jobs
            )

        self.jobs = None

    def get_online_solves(self) -> bool:
        if self.points is None:
            raise ValueError("Points not initialized")
        if self.jobs is None:
            raise ValueError("Jobs not initialized")

        self.setup_online_solves()

        results = _neos.get_solves(
            client=self.neos,
            job_list=self.jobs,
            points_list=self.points
        )

        if all([_ is not None for _ in results]):
            self.points = results
            self.is_routed = True
            self.jobs = None
            return True

        return False

    def get_online_solves_blocking(self, delay_minutes=0.25, logging=True):
        if self.points is None:
            raise ValueError("Points not initialized")
        if self.jobs is None:
            raise ValueError("Jobs not initialized")

        self.setup_online_solves()

        results = _neos.get_solves_blocking(
            client=self.neos,
            job_list=self.jobs,
            points_list=self.points,
            delay_minutes=delay_minutes,
            logging=logging
        )

        self.points = results
        self.is_routed = True
        self.jobs = None

    def offline_solves(self, time_limit_minutes=60, symmetric=True, logging=True, verbose=False):
        if self.points is None:
            raise ValueError("Points not initialized")

        self.cancel_online_solves()

        results = _tsp.heuristic_solves(
            points_list=self.points,
            time_limit_minutes=time_limit_minutes,
            symmetric=symmetric,
            logging=logging,
            verbose=verbose
        )

        if all([_ is not None for _ in results]):
            self.points = results
            self.is_routed = True
            return True

        return False

    def draw(self, scale=1, closed=True, subpixels=8):
        if self.points is None:
            raise ValueError("Points not initialized")
        if self.factors is None:
            raise ValueError("Factors not initialized")
        if not self.is_routed:
            raise ValueError("Not routed yet")

        match self.mode:
            case ColorMode.CMYK:
                return _draw_cmyk_routes(
                    cmyk_points=self.points,
                    cmyk_factors=self.factors,
                    size=self.size,
                    line_width=self.line_width,
                    scale=scale,
                    closed=closed,
                    subpixels=subpixels
                )
            case ColorMode.RGB:
                return _draw_rgb_routes(
                    rgb_points=self.points,
                    rgb_factors=self.factors,
                    size=self.size,
                    line_width=self.line_width,
                    scale=scale,
                    closed=closed,
                    subpixels=subpixels
                )
            case ColorMode.GRAYSCALE:
                return _draw_route(
                    points=self.points[0],
                    factors=self.factors[0],
                    size=self.size,
                    line_width=self.line_width,
                    scale=scale,
                    closed=closed,
                    background=self.background,
                    foreground=self.foreground,
                    subpixels=subpixels
                )


def load(filename) -> TspStudio:
    obj = _load_json(filename)

    obj["mode"] = ColorMode(obj["mode"])
    if obj["points"] is not None:
        obj["points"] = _array_to_ndarray_2d(obj["points"])
    obj["background"] = tuple(obj["background"])
    obj["foreground"] = tuple(obj["foreground"])

    return TspStudio(
        mode=obj["mode"],
        image=obj["image"],
        num_points=obj["num_points"],
        line_width=obj["line_width"],
        white_threshold=obj["white_threshold"],
        invert=obj["invert"],
        background=obj["background"],
        foreground=obj["foreground"],
        points=obj["points"],
        is_routed=obj["is_routed"],
        jobs=obj["jobs"]
    )


def save(filename, studio, round_places=2):
    obj = studio.data

    obj["mode"] = obj["mode"].value
    if obj["points"] is not None:
        obj["points"] = _ndarray_to_array_2d(
            array=obj["points"],
            round_places=round_places
        )

    _save_json(filename, obj, indent=None)



import base64
from io import BytesIO

import numpy as np
from PIL import Image


def split_cmyk(rgb_array, invert=False, threshhold=1):
    data = rgb_array.astype(float) / 255
    threshold = threshhold / 255

    channel_max = data.max(2)
    channel_max[channel_max < threshold] = threshold

    k = 1 - channel_max
    c = (1 - data[:, :, 0] - k) / channel_max
    m = (1 - data[:, :, 1] - k) / channel_max
    y = (1 - data[:, :, 2] - k) / channel_max

    result = 1 - np.array([c, m, y, k])

    result = tuple([(_ * 255).round().astype(int) for _ in result])
    if invert:
        result = invert_array_multi(result)

    return result


def split_rgb(rgb_array, invert=False):
    r = rgb_array[:, :, 0]
    g = rgb_array[:, :, 1]
    b = rgb_array[:, :, 2]

    result = (r, g, b)
    if invert:
        result = invert_array_multi(result)

    return result


def luminance(rgb):
    r, g, b = rgb

    return int(round((0.299 * r) + (0.587 * g) + (0.114 * b)))


def rgb_to_grayscale(rgb_array, invert=False):
    result = np.array([[luminance(__) for __ in _] for _ in rgb_array])
    if invert:
        result = invert_array_multi(result)

    return result


def invert_array(grayscale_array):
    return 255 - grayscale_array


def invert_array_multi(grayscale_arrays):
    return [invert_array(_) for _ in grayscale_arrays]


def image_to_base64(image, format="JPEG"):
    buffered = BytesIO()
    image.save(buffered, format=format)

    return base64.b64encode(buffered.getvalue()).decode("utf-8")


def base64_to_image(base64_string):
    image_bytes = BytesIO(base64.b64decode(base64_string))

    return Image.open(image_bytes)
